/// <reference types="node" />
import { Stats } from "fs";
import { ReadOptionHooks, ReadOptions, StreamRecord } from "../../types";
import { RStreamsSdk } from "../../../index";
import { S3ClientConfig } from "@aws-sdk/client-s3";
export interface ReadHooksParams {
    /**
     * Configuration for the AWS S3 Client used in the download threads
     */
    awsS3Config?: S3ClientConfig;
    /**
     * Amount of space available to use when downloading S3 files.
     * @default 419430400 (400 MB)
     */
    availableDiskSpace?: number;
    /**
     * Can be undefined|1
     *
     * Determines how files are merged to local disk
     */
    mergeFileVersion?: number;
    /**
     * Target file size when merging S3 files to local disk
     */
    mergeFileSize?: number;
    /**
     * Number of threads to use to download S3 files
     */
    downloadThreads?: number;
    /**
     * number of threads to use to parse string json into objects
     */
    parseThreads?: number;
    /**
     * @ignore
    */
    payloadAtEnd?: boolean;
    /**
     * @ignore
    */
    unzipFiles?: boolean;
    /**
     * Save the local files rather then deleting them after being consumed
     * @default false
    */
    saveFiles?: boolean;
    /**
     * Directory used when downloading files
     * @default /tmp/rstreams
     */
    tmpDir?: string;
    /**
     * Path to the download task code
     *
     * @default internal code
     */
    downloadTaskPath?: string;
    /**
     * Path to the parse task code
     *
     * @default internal code
     */
    parseTaskPath?: string;
    /**
     * Weather or not to enable parallel parsing
     */
    parallelParse?: boolean;
    /**
     * Parse Gzip content in threads too.
     * This flag is not fully supported yet
     * @ignore
     */
    parallelParseGzip?: boolean;
    /**
     * Settings for the Parser task
     */
    parseTaskParser?: {
        /**
         * Custom options to pass to the parser
         */
        opts?: any;
        /**
         * Name of the parser or Path to the parser factory file
         */
        parser?: string;
        /**
         * Data size used to send parsed data between threads
         */
        bufferSize?: number;
    };
}
interface StatsPlus extends Stats {
    fullpath: string;
}
export declare function getAllFiles(dirPath: string, arrayOfFiles?: StatsPlus[]): StatsPlus[];
interface LocalFileStreamRecordS3 {
    bucket: string;
    key: string;
    start: string;
    end: string;
    gzipSize: number;
    writeStartEid: boolean;
}
interface LocalFileStreamRecord extends StreamRecord {
    s3Like?: boolean;
    filePrefix?: string;
    localFile: {
        id: number;
        filePath: string;
        readyPromise: PromiseResolver<void>;
        unlinkOnStreamClose?: string[];
        data?: any;
    };
    gzip?: string;
    s3?: LocalFileStreamRecordS3 | LocalFileStreamRecordS3[];
    offsets?: any[];
    s3Parts?: LocalFileStreamRecord[];
    mergeS3GzipSize?: number;
}
interface ExtraHooks<SR extends StreamRecord> {
    onGetEventsAddDownload: (StreamRecords: SR[]) => void;
    onGetEventsMerge: (StreamRecords: SR[]) => void | SR[];
    getExtraMetaData: () => {
        availableDiskSpace: number;
        startingDiskSpace: number;
        enddingDiskSpace: number;
    };
}
/**
 * Creates read hooks that use worker threads for downloading and parsing
 * @param settings
 * @returns
 */
export declare function createFastS3ReadHooks(settings: ReadHooksParams): ReadOptionHooks<LocalFileStreamRecord> & ExtraHooks<LocalFileStreamRecord>;
export interface ExtraConfig {
    tmpDir?: string;
    awsS3Config?: S3ClientConfig;
}
/**
 * Addes configured read hooks to the read options
 * @param readOpts
 * @param partialHookSettings
 * @param extraConfig
 * @returns
 */
export declare function addReadHooks<T>(readOpts: ReadOptions<T>, partialHookSettings?: Partial<ReadHooksParams>, extraConfig?: RStreamsSdk | ExtraConfig): ReadOptions<T>;
/**
 *
 * @param settings
 * @param partialHookSettings
 * @param extraConfig
 * @returns
 */
export declare function determineReadHooks<T>(settings: ReadOptions<T>, partialHookSettings?: Partial<ReadHooksParams>, extraConfig?: RStreamsSdk | ExtraConfig): Partial<ReadOptions<T>>;
interface PromiseResolver<T> extends Promise<T> {
    resolve: () => void;
    reject: (err?: any) => void;
}
export {};
